<?php
namespace modules\tariff;


use components\database\db;
use core\array_collection;
use core\env;
use core\response;
use modules\pbx\models\operator;
use modules\tariff\models\carrier_rates as carrier_rates_model;
use modules\tariff\models\customer_rates as customer_rates_model;

class dial_codes_importer{
	private static $_csv_headers = [
		'group_name',
		'code',
		'call_cost',
		'cost_interval',
		'code_description',
		'dial_prefix'
	];

	private static $_required_headers = [
		'group_name',
		'code',
		'call_cost'
	];

	private static $_positive_aliases = [ '1', 'yes', 'true', 'y' ];
	private static $_toggle_fields = [];

	private static $_log = [];
	private static $_imported_count = 0;
	private static $_skipped_count = 0;
	private static $_processed_count = 0;
	private static $_mode;
	private static $_dial_codes_regex = '/[0-9XNZ\[\]\-*+|]+$/i';
	private static $_to_import = [];
	private static $_categories = [];
	private static $_operators = [];

	public static function get_format($prefix_name){
		header('Content-type: text/csv');
		header('Content-Disposition: attachment; filename="'.$prefix_name.'_rates.csv"');
		header('Cache-Control: private, max-age=0, must-revalidate');
		header('Pragma: public');
		ini_set('html_errors', false);
		$columns = array_flip(self::$_csv_headers);
		foreach($columns as $column => $value) {
			$columns[$column] = '"' . trans("import_rates.{$column}.description") . '"';
		}

		$descriptions = implode(',', $columns);
		$headers = implode(',', array_keys($columns));
		echo implode("\r\n", [$headers]);
		exit;
	}

	public static function import($mode){
		self::$_mode = $mode;
		self::_populate_categories();
		self::_populate_operators();

		try{
			$csvContent = env::request_post('file_content', false);
			if(preg_match('/^(FFFE|FEFF)/i', bin2hex($csvContent))) { // UTF-16 BOM found
				$csvContent = iconv('utf-16', 'utf-8', $csvContent); // Convert to UTF-8
			}

			$csvContent = array_filter(preg_split('/[\r\n]+/', $csvContent), 'strlen'); //Delete empty lines
			$rows = array_map('str_getcsv', $csvContent); //Give CSV Format
			$importInfo = self::_process_data($rows);
			response::custom('imported', $importInfo);
		}catch (\Exception $e){
			response::notification('error', 'Exception', sprintf(
				'error: %s, line: %s, file: %s',
				$e->getMessage(),
				$e->getLine(),
				$e->getFile()
			));
		}
	}

	private static function _process_data(array $rows){
		$header = array_map('trim', $rows[0]);
		unset($rows[0]);
		self::validateCSVHeaders($header);
		$rows = new \ArrayObject($rows);

		$line = 2;
		foreach ($rows as $index => $row){
			$row = array_map('trim', $row);
			$row = array_combine($header, $row);
			self::_filter($row, array_flip($header));
			self::_set_defaults($row);
			self::$_processed_count += 1;

			if(!preg_match(self::$_dial_codes_regex, $row->code)){
				self::$_skipped_count += 1;
				self::_log(trans('customer_rates.invalid',['p' => $row->code]), $line, true);
				$line++;
				continue;
			}

			$row->code = strtoupper($row->code);
			$k = str_replace(['X', 'Z', 'N'], ['[0-9]', '[1-9]', '[2-9]'], $row->code);
			$row->group_name = utf8_encode($row->group_name);

			//Default Data
			$row->dial_prefix = preg_match('/[*#+0-9]+$/i', $row->dial_prefix) ? $row->dial_prefix : '';
			$row->call_cost = is_numeric($row->call_cost) ? $row->call_cost : 0;
			$row->cost_interval =  filter_var($row->cost_interval, \FILTER_VALIDATE_INT) !== false ? $row->cost_interval : 60;

			$k = "{$row->dial_prefix}{$k}";
			$data = [
				'prefix' => $row->code,
				'description' => utf8_encode($row->code_description),
				'strip' => $row->dial_prefix,
				'cost' => $row->call_cost,
				'cost_interval' => $row->cost_interval,
				'sort' => 0,
			];

			self::_add_dial_code_to_import($row->group_name, $k, (object) $data, $line);
			$line++;
		}

		self::_import_rates();

		return [
			'processed' => self::$_processed_count,
			'skipped' => self::$_skipped_count,
			'updated' => 0,
			'added' => self::$_imported_count,
			'removed' => 0
		];
	}

	private static function validateCSVHeaders($header){
		$csv_header = array_map('trim', $header);
		if(count(array_diff(self::$_required_headers, $csv_header)) != 0) {
			response::notification( 'error', trans(self::$_mode === 'carrier' ? 'carrier_rates': 'customer_rates'), trans('invalid_csv_format'));
		}

		return $csv_header;
	}

	private static function _import_rates(){
		$rows = array_collection::make(self::$_to_import);
		db::start_transaction();
		$rows->each(function ($data, $group_name){
			if(self::$_mode === 'carrier'){
				self::_save_carrier_rates($group_name, $data);
			}else{
				$id = array_get($group_name, self::$_categories);
				self::_save_customer_rates($group_name, $data, $id);
			}

		});
		db::end_transaction();
	}

	private static function _save_carrier_rates($group_name, array $data){
		$operator_id = env::request_post('operator_id');
		$operator_name = array_get($operator_id, self::$_operators, null);
		$description = "{$group_name} - {$operator_name}";
		$id = array_get($description, self::$_categories);
		$model = new carrier_rates_model($id);

		$duplicated_prefixes = carrier_rates_model::find_prefixes(
			array_keys($data),
			$operator_id,
			'out',
			$id
		);

		if($duplicated_prefixes){ //Remove duplicated Prefixes From This Rates
			$data = array_diff_key($data, $duplicated_prefixes);
			self::_log(trans_choice(
				'rates_importer.duplicated_dial_code_on_importation',
				count($duplicated_prefixes),
				['code' => implode(',', $duplicated_prefixes) ,'group' => $group_name]
			), 0, true);
			self::$_skipped_count += count($duplicated_prefixes);
		}

		if(!$model->carrier_rate_id || ($model->carrier_rate_id && $operator_id != $model->operator_id)){
			$model->operator_id = $operator_id;
			$model->prefix_category_id = null;
			$model->description = $description;
			$model->strip = null;
			$model->free_time = 0;
			$model->connect_cost = 0;
			$model->cost = 0;
			$model->cost_interval = 60;
			$model->call_direction = 'out'; // By now we will support only rates for outgoing calls
		}

		$model->save($data);
		self::$_imported_count += count($data);
	}

	private static function _save_customer_rates($group_name, array $data, $id = null){
		$model = new customer_rates_model($id);

		$duplicated_prefixes = customer_rates_model::find_prefixes(array_keys($data), 'out', $id);
		if($duplicated_prefixes){ //Remove duplicated Prefixes From This Rates
			$data = array_diff_key($data, $duplicated_prefixes);
			self::_log(trans_choice(
				'rates_importer.duplicated_dial_code_on_importation',
				count($duplicated_prefixes),
				['code' => implode(',', $duplicated_prefixes) ,'group' => $group_name]
			), 0, true);
			self::$_skipped_count += count($duplicated_prefixes);
		}

		if(!$model->customer_rate_id){ //If not exists fill with default values
			$model->description = $group_name;
			$model->prefix_category_id = null;
			$model->strip = null;
			$model->free_time = 0;
			$model->connect_cost = 0;
			$model->cost = 0;
			$model->cost_interval = 60;
			$model->call_direction = 'out'; // By now we will support only rates for outgoing calls
		}

		$model->save($data);
		self::$_imported_count += count($data);

	}

	private static function _filter(&$row, $headers){
		$filtered = (object) [];

		foreach (self::$_csv_headers as $header) {
			$filtered->$header = null;
			if(array_key_exists($header, $headers)) {
				$value = trim($row[$header] ?? '');
				if(array_key_exists($header, array_flip(self::$_toggle_fields))) {
					if(strlen($value ?? '') > 0) {
						$filtered->$header = in_array(strtolower($value), self::$_positive_aliases) ? true : false;
					}
				} else {
					$filtered->$header = $value;
				}
			}
		}

		$row = $filtered;
	}

	private static function _set_defaults(&$row){
		if(!self::_column_exists($row, 'cost_interval')){
			$row->cost_interval = 60;
		}

		if(!self::_column_exists($row, 'code_description')){
			$row->code_description = '';
		}

		if(!self::_column_exists($row, 'dial_prefix')){
			$row->dial_prefix = '';
		}
	}

	private static function _column_exists($row, $column_name){
		return property_exists($row, $column_name) && (strlen($row->$column_name ?? '') > 0);
	}

	private static function _log($message, $line, $error = false){
		self::$_log[] = (object)[
			'error' => $error,
			'message' => trans('rates_importer.log_message', ['row' => $line, 'message' => $message])
		];
	}

	private static function _add_dial_code_to_import($group, $k, $data, $line){
		if(!array_key_exists($group, self::$_to_import)){
			self::$_to_import[$group] = [];
		}

		if(!array_key_exists($k, self::$_to_import[$group])){
			self::$_to_import[$group][$k] = $data;
		}else{
			self::$_skipped_count += 1;
			self::_log(trans_choice('rates_importer.duplicated_dial_code_on_importation', 1, ['code' => $data->prefix ,'group' => $group]), $line, true);
		}

	}

	private static function _populate_categories() {
		if(self::$_mode === 'carrier'){
			foreach (carrier_rates_model::all() as $row){
				self::$_categories[$row->description] = $row->carrier_rate_id;
			}
		}else{
			foreach (customer_rates_model::all() as $row){
				self::$_categories[$row->description] = $row->customer_rate_id;
			}
		}
	}

	private static function _populate_operators(){
		foreach (operator::all() as $operator){
			self::$_operators[$operator->operator_id] = $operator->name;
		}
	}
}